/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <netdb.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/socket.h>
#include <time.h>

#include "dataio/TCPSocketDataIO.h"
#include "iogateway/PortableMessageIOGateway.h"
#include "reflector/StorageReflectConstants.h"
#include "netutil/NetworkUtilityFunctions.h"

#define TEST(x) if ((x) != B_NO_ERROR) printf("Test failed, line %i\n",__LINE__)

// This is a text based test client for the muscled server.  It is useful for testing
// the server, and could possibly be useful for other things, I don't know.
int main(int argc, char ** argv)
{
#ifdef __BEOS__
   printf("Warning:  This program doesn't run nicely under BeOS.  Try testreflectclient instead!\n");
#endif

   char * hostName = "localhost";
   int port = 0;
   if (argc > 1) hostName = argv[1];
   if (argc > 2) port = atoi(argv[2]);
   if (port == 0) port = 2960;

   int s = Connect(hostName, port, "portablereflectclient");
   if (s < 0) exit(10);

   PortableMessageIOGateway gw(new TCPSocketDataIO(s, false));
   char text[1000] = "";
   bool keepGoing = true;
   fd_set readSet, writeSet;
   while(keepGoing)
   {
      FD_ZERO(&readSet);
      FD_ZERO(&writeSet);

      int maxfd = s;
      FD_SET(s, &readSet);
      struct timeval * timeout = NULL;
      struct timeval poll = {0, 0};

      if (gw.HasBytesToOutput()) FD_SET(s, &writeSet);

#ifdef __BEOS__
      // BeOS can't select() on stdin (yet), so just make the user press enter or whatever
      printf("Enter something => "); 
      fflush(stdout);
      fgets(text, sizeof(text), stdin);
      char * ret = strchr(text, '\n'); if (ret) *ret = '\0';
      timeout = &poll;  // prohibit blocking in select()
#else
      if (STDIN_FILENO > maxfd) maxfd = STDIN_FILENO;
      FD_SET(STDIN_FILENO, &readSet);
#endif

      while(keepGoing) 
      {
         if (select(maxfd+1, &readSet, &writeSet, NULL, timeout) < 0) printf("portablereflectclient: select() failed!\n");

         timeout = &poll;  // only block the first time (Linux) or not at all (BeOS)

#ifndef __BEOS__
         if (FD_ISSET(STDIN_FILENO, &readSet)) 
         {
            fgets(text, sizeof(text), stdin);
            char * ret = strchr(text, '\n'); if (ret) *ret = '\0';
         }
#endif

         if (text[0])
         {
            printf("You typed: [%s]\n",text);
            bool send = true;
            PortableMessage * msg = new PortableMessage;
            PortableMessageRef ref(msg, NULL, NULL);

            switch(text[0])
            {
               case 'm':
                  msg->what = MAKETYPE("umsg");
                  msg->AddString(PR_NAME_KEYS, &text[2]);
                  msg->AddString("info", "This is a user message");
               break;

               case 's':
                  msg->what = PR_COMMAND_SETDATA;
                  msg->AddMessage(&text[2], PortableMessage(MAKETYPE("HELO"))); 
               break;
   
               case 'k':
                  msg->what = PR_COMMAND_KICK;
                  msg->AddString(PR_NAME_KEYS, &text[2]);
               break;

               case 'b':
                  msg->what = PR_COMMAND_ADDBANS;
                  msg->AddString(PR_NAME_KEYS, &text[2]);
               break;

               case 'B':
                  msg->what = PR_COMMAND_REMOVEBANS;
                  msg->AddString(PR_NAME_KEYS, &text[2]);
               break;

               case 'g':
                  msg->what = PR_COMMAND_GETDATA;
                  msg->AddString(PR_NAME_KEYS, &text[2]);
               break;
   
               case 'q':
                  keepGoing = send = false;
               break;
   
               case 'p':
                  msg->what = PR_COMMAND_SETPARAMETERS;
                  msg->AddString(&text[2], "");
               break;
   
               case 'P':
                  msg->what = PR_COMMAND_GETPARAMETERS;
               break;
   
               case 'd':
                  msg->what = PR_COMMAND_REMOVEDATA;
                  msg->AddString(PR_NAME_KEYS, &text[2]);
               break;
   
               case 'D':
                  msg->what = PR_COMMAND_REMOVEPARAMETERS;
                  msg->AddString(PR_NAME_KEYS, &text[2]);
               break;
   
               case 't':
               {
                  // test all data types
                  msg->what = 1234;
                  msg->AddString("String", "this is a string");
                  msg->AddInt8("Int8", 123);
                  msg->AddInt8("-Int8", -123);
                  msg->AddInt16("Int16", 1234);
                  msg->AddInt16("-Int16", -1234);
                  msg->AddInt32("Int32", 12345);
                  msg->AddInt32("-Int32", -12345);
                  msg->AddInt64("Int64", 123456789);
                  msg->AddInt64("-Int64", -123456789);
                  msg->AddBool("Bool", true);
                  msg->AddBool("-Bool", false);
                  msg->AddFloat("Float", 1234.56789f);
                  msg->AddFloat("-Float", -1234.56789f);
                  msg->AddDouble("Double", 1234.56789);
                  msg->AddDouble("-Double", -1234.56789);
                  msg->AddPointer("Pointer", msg);
                  msg->AddFlat("Flat", *msg);
                  char data[] = "This is some data";
                  msg->AddData("Flat", B_RAW_TYPE, data, sizeof(data));
               }
               break;

               break;

               default:
                  printf("Sorry, wot?\n");
                  send = false;
               break;
            }
   
            if (send) 
            {
               printf("Sending message...\n");
//             ref.GetItemPointer()->PrintToStream();
               gw.AddOutgoingMessage(ref);
            }
            text[0] = '\0';
         }
   
         bool reading = FD_ISSET(s, &readSet);
         bool writing = FD_ISSET(s, &writeSet);
         bool writeError = ((writing)&&(gw.DoOutput() != B_NO_ERROR));
         bool readError  = ((reading)&&(gw.DoInput() != B_NO_ERROR));
         if ((readError)||(writeError))
         {
            printf("Connection closed, exiting.\n");
            closesocket(s);
            keepGoing = false;
         }

         PortableMessageRef incoming;
         while(gw.GetNextIncomingMessage(incoming) == B_NO_ERROR)
         {
            printf("Heard message from server:-----------------------------------\n");
            incoming()->PrintToStream();
            printf("-------------------------------------------------------------\n");
         }

         if ((reading == false)&&(writing == false)) break;

         FD_ZERO(&readSet);
         FD_ZERO(&writeSet);
         FD_SET(s, &readSet);
         if (gw.HasBytesToOutput()) FD_SET(s, &writeSet);
      }
   } 
   closesocket(s);

   if (gw.HasBytesToOutput())
   {
      printf("Waiting for all pending messages to be sent...\n");
      while((gw.HasBytesToOutput())&&(gw.DoOutput() == B_NO_ERROR)) {printf ("."); fflush(stdout);}
   }
   printf("\n\nBye!\n");
   return 0;
}
