/* This file is Copyright 2000 Level Control Systems.  See the included LICENSE.txt file for details. */  

#include <stdio.h>
#include <stdarg.h>

#include "hashtable/PortableHashtable.h"
#include "string/PortableString.h"

int _state = 0;

static void bomb(const char * fmt, ...);
void bomb(const char * fmt, ...)
{
   va_list va;
   va_start(va, fmt);
   vprintf(fmt, va);
   va_end(va);
   printf("EXITING DUE TO ERROR (state = %i)!\n", _state);
   exit(10);
}

uint32 SillyHashFunc(const uint32 & val);
uint32 SillyHashFunc(const uint32 & val)
{
   return val;
}

// This program exercises the PortableHashtable class.
void main(void) 
{
   PortableHashtable<PortableString, PortableString> table(PortableStringHashFunc);
   {
   table.Put(PortableString("Hello"), PortableString("World"));

   table.Put(PortableString("Peanut Butter"), PortableString("Jelly"));
   table.Put(PortableString("Ham"), PortableString("Eggs"));
   table.Put(PortableString("Pork"), PortableString("Beans"));
   table.Put(PortableString("Slash"), PortableString("Dot"));
   table.Put(PortableString("Data"), PortableString("Mining"));  
   table.Put(PortableString("TestDouble"), PortableString("Play"));

   printf("String table has %i entries in it...\n", table.Size());

   PortableString lookup;
   if (table.Get(PortableString("Hello"), lookup)) printf("Hello -> %s\n", lookup());
                                              else printf("Lookup 1 failed.\n");
   if (table.Get(PortableString("Peanut Butter"), lookup)) printf("Peanut Butter -> %s\n", lookup());
                                              else printf("Lookup 2 failed.\n");


   PortableHashtableIterator<PortableString, PortableString> st = table.GetIterator();
   PortableString nextKeyString;
   PortableString nextValueString;

   while(st.GetNextKey(nextKeyString))
   {
      bool ret = table.Get(nextKeyString, nextValueString);
      printf("%i %s: %s -> %s\n", st.HasMoreKeys(), ret ? "OK" : "ERROR", nextKeyString(), nextValueString());
   }

   PortableHashtable<uint32, const char *> sillyTable(SillyHashFunc);
   sillyTable.Put(15, "Fifteen");
   sillyTable.Put(100, "One Hundred");
   sillyTable.Put(150, "One Hundred and Fifty");
   sillyTable.Put(200, "Two Hundred");

   const char * temp;
   sillyTable.Get(100, temp);
   sillyTable.Get(101, temp); // will fail
   printf("100 -> %s\n",temp);

   printf("Entries in sillyTable:\n");
   PortableHashtableIterator<uint32, const char *> it = sillyTable.GetIterator();

   uint32 nextKey;
   const char * nextValue;
   while(it.GetNextKey(nextKey))
   {
      bool ret = sillyTable.Get(nextKey, nextValue);
      printf("%i %s: %u -> %s\n", it.HasMoreKeys(), ret ? "OK" : "ERROR", nextKey, nextValue);
   }
   printf("XXX %i\n", it.HasMoreKeys());

   printf("... and by value:\n");
   while(it.GetNextValue(nextValue))
   {
      printf("%i value: %s\n", it.HasMoreValues(), nextValue);
   }
   printf("YYY %i\n", it.HasMoreValues());

/*
   printf("Testing concurrent modification detection (this should cause an assertion failure)\n");
   PortableHashtableIterator<uint32, const char *> it2 = sillyTable.GetIterator();
   sillyTable.Put(250, "Crash me baby!");
   it2.GetNextKey(nextKey); 
   printf("You shouldn't ever see this message!\n");
*/
   }
   table.Clear();   

   printf("Begin torture test!\n");
   _state = 4;
   bool fastClear = false;
   PortableHashtable<PortableString, uint32> t(PortableStringHashFunc);
   while(true)
   {
      for (int numEntries=1; numEntries < 100000; numEntries++)
      {
         printf("Populating table with %i entries...\n", numEntries);
         _state = 5;
         {
            for(int i=0; i<numEntries; i++)
            {
               char temp[300];
               sprintf(temp, "%i", i);
               t.Put(temp, i);
            }
         }

         //printf("Checking that all entries are still there...\n");
         _state = 6;
         {
            if (t.Size() != numEntries) bomb("ERROR, WRONG SIZE %i vs %i!\n", t.Size(), numEntries);
            for (int i=numEntries-1; i>=0; i--)
            {
               char temp[300];
               sprintf(temp, "%i", i);
               uint32 tv;
               if (t.Get(temp, tv) == false) bomb("ERROR, MISSING KEY [%s]\n", temp);
               if (tv != i) bomb("ERROR, WRONG KEY %s != %lu!\n", temp, tv);
            }
         }
   
         //printf("Iterating through table...\n");
         _state = 7;
         {
            uint32 sum = 0; 
            for (int i=0; i<numEntries; i++) sum += i;

            uint32 count = 0, checkSum = 0;
            PortableHashtableIterator<PortableString, uint32> iter = t.GetIterator();
            PortableString nextKey;
            while(iter.GetNextKey(nextKey)) 
            {
               count++;
               uint32 temp;
               (void) iter.GetNextValue(temp);
               checkSum += temp;
            }
            if (count != numEntries) bomb("ERROR: Count mismatch %i vs %i!\n", count, numEntries);
            if (checkSum != sum)     bomb("ERROR: Sum mismatch %i vs %i!\n", sum, checkSum);
         }

         //printf("Removing the second half of the entries...\n");
         _state = 8;
         int half = numEntries/2;
         {
            for (int i=half; i<numEntries; i++)
            {
               char temp[300];
               sprintf(temp, "%i", i);
               uint32 tv;
               if (t.Remove(temp, tv) == false) bomb("ERROR, MISSING REMOVE KEY [%s]\n", temp);
               if (tv != i) bomb("ERROR, REMOVE WAS WRONG VALUE %lu\n", tv);
            }
         }

         //printf("Iterating over only first half...\n");
         _state = 9;
         {
            uint32 sum = 0; 
            for (int i=0; i<half; i++) sum += i;

            uint32 count = 0, checkSum = 0;
            PortableHashtableIterator<PortableString, uint32> iter = t.GetIterator();
            PortableString nextKey;
            while(iter.GetNextKey(nextKey)) 
            {
               count++;
               uint32 temp;
               (void) iter.GetNextValue(temp);
               checkSum += temp;
            }
            if (count != half) bomb("ERROR: Count mismatch %i vs %i!\n", count, numEntries);
            if (checkSum != sum)     bomb("ERROR: Sum mismatch %i vs %i!\n", sum, checkSum);
         }

         //printf("Clearing Table (%s)\n", fastClear ? "Quickly" : "Slowly"); 
         _state = 10;
         if (fastClear) t.Clear();
         else
         {
            for (int i=0; i<half; i++)
            {
               char temp[300];
               sprintf(temp, "%i", i);
               uint32 tv;
               if (t.Remove(temp, tv) == false) bomb("ERROR, MISSING REMOVE KEY [%s]\n", temp);
               if (tv != i) bomb("ERROR, REMOVE WAS WRONG VALUE %lu\n", tv);
            }
         }

         PortableHashtableIterator<PortableString, uint32> paranoia = t.GetIterator();
         if (paranoia.GetNextKey()) bomb("ERROR, ITERATOR CONTAINED ITEMS AFTER CLEAR!\n");

         if (t.Size() != 0) bomb("ERROR, SIZE WAS NON-ZERO (%lu) AFTER CLEAR!\n", t.Size());
         fastClear = !fastClear;
      }
      printf("Finished torture test successfully, testing again...\n");
   }
}
